#include "helper.h"

long time_stamp = 0;

// Pins for the shift register
int latchPin = 11;
int clockPin = 13;
int dataPin = 10;

byte leds = 0;
int level = 0;

// Initialising the TouchScreen object
TouchScreen ts = TouchScreen(XP, YP, XM, YM, 300);
int isDay = 0;

int buttonPin = PB4;
const int buzzer = A5;
int tempo = 258;
int lastPress = -100;
bool started = false;

// Every 4 seconds, increase the difficulty
ISR(TIMER1_COMPA_vect)
{
  if (!started) return;
  if (time_stamp == 16) {
    time_stamp = 0;
    level++;
  } else time_stamp++;
}

int flappyColor = BLUE;
int colorChanged = 0;

// If the button is pressed, change the bird's color
ISR (PCINT0_vect)
{    
  if (millis() - time_stamp > 300) {
    time_stamp = millis();
    colorChanged = 1;
    if (flappyColor == BLUE)
      flappyColor = YELLOW;
    else if (flappyColor == YELLOW)
      flappyColor = RED;
    else if (flappyColor == RED)
      flappyColor = CYAN;
    else if (flappyColor == CYAN)
      flappyColor = PURPLE;
    else if (flappyColor == PURPLE)
      flappyColor = GREEN;
    else if (flappyColor == GREEN)
      flappyColor = ORANGE;
    else if (flappyColor == ORANGE)
      flappyColor = MAGENTA;
    else if (flappyColor == MAGENTA)
      flappyColor = WHITE;
    else if (flappyColor == WHITE)
      flappyColor = BLUE;
  }
}

//The songs that are played when a bird passes through a tube;
//or when the bird crashes.
int coin_melody[] = {
  NOTE_A6,8, NOTE_DS6, -8
};

int fail_melody[] = {
  NOTE_C6,8, NOTE_A6, 4, NOTE_DS6, -4
};

int coin_notes = sizeof(coin_melody) / sizeof(coin_melody[0]) / 2;
int fail_notes = sizeof(fail_melody) / sizeof(fail_melody[0]) / 2;

int wholenote = (60000 * 4) / tempo;

int divider = 0, noteDuration = 0;
int flying = 0;
int animationCooldown = 0;
bool lost = false;
int score = 0;
int high_score = -1;
int pipe_space;

// Updates the leds from the shift register
void updateShiftRegister()
{
   digitalWrite(latchPin, LOW);
   shiftOut(dataPin, clockPin, LSBFIRST, leds);
   digitalWrite(latchPin, HIGH);
}

/* Draws our precious little bird, using the color from the parameter.
 x and y represent the position of the bird
 sx and sy are scale factors */
void drawFlappy(int x, int y, int color, float sx, float sy) {

    tft.fillRoundRect(x + 5 * sx, y + 10 * sy, 17 * sx, 17 * sy, 6, color);
    tft.fillRoundRect(x - 3 * sx, y + 5 * sy, 15 * sx, 15 * sy, 5, color);

    tft.fillRoundRect(x + 7 * sx, y + 14 * sy, 10 * sx, 10 * sy, 2, WHITE);
    tft.fillRect(x + 12 * sx, y + 18 * sy, 4 * sx, 4 * sy, BLACK);
    tft.fillRect(x + 10 * sx, y + 23 * sy, 4 * sx, 8 * sy, YELLOW);

    tft.fillRect(x - 6 * sx, y + 5 * sy, 11 * sx, 6 * sy, color);
    tft.fillRect(x - 3 * sx, y + 2 * sy, 6 * sx, 6 * sy, color);

    tft.fillRoundRect(x - 3 * sx, y + 10 * sy, 8 * sx, 9 * sy, 2, WHITE);

    tft.fillRect(x + 5 * sx, y + 2 * sy, 8 * sx, 10 * sy, color);
    tft.fillRect(x + 5 * sx, y + 2 * sy, 10 * sx, 6 * sy, color);
    tft.fillRect(x + 7 * sx, y, 10 * sx, 6 * sy, color);

}

// Deletes the bird by drawing a rectangle over it
void clearFlappy(int x, int y, float sx, float sy) {
    if (isDay)
      tft.fillRect(x - 6 * sx, y, 30 * sx, 30 * sy, DAY);
    else
      tft.fillRect(x - 6 * sx, y, 30 * sx, 30 * sy, NIGHT);
}

// Checks if the TouchScreen is pressed
bool isPressed() {

  TSPoint p = ts.getPoint();
  pinMode(XM, OUTPUT);
  pinMode(YP, OUTPUT);
  int crt_press = millis();
  if (crt_press < lastPress + 100) {
    return false;
  }
  lastPress = crt_press;
  if (p.z > ts.pressureThreshhold) {
     return true;
  }

  return false;
}

// Draws the text for the start menu
void drawStartText() {
  tft.setRotation(1);
  tft.setCursor(100, 140);
  tft.setTextColor(YELLOW);  tft.setTextSize(4);
  tft.println("Tap to start");
  tft.setRotation(0);
}

/* Draws the text when the bird crashes, including
 *  the current score and the high score
 */
void drawGameOver() {
  tft.setRotation(1);
  
  tft.setCursor(100, 80);
  tft.setTextColor(YELLOW);  tft.setTextSize(4);
  tft.println("Game Over!");
  
  tft.setCursor(100, 140);
  tft.setTextColor(YELLOW);  tft.setTextSize(4);
  tft.println("Score: " + String(score));

  tft.setCursor(100, 200);
  tft.setTextColor(YELLOW);  tft.setTextSize(4);
  tft.println("High score: " + String(high_score));

  tft.setRotation(0);
  playFailMelody(fail_melody, fail_notes);
}

// Draws the night environment
void drawNight() {
  tft.fillScreen(NIGHT);
  tft.fillRect(0, 0, 16, 500, 0x5ACB);
  if (!started) {
    drawBitmap(255, 410, moon, 60, 60,GRAY);
    drawBitmap(245, 30, stars, 102, 80,GRAY);
    drawBitmap(245, 0, stars, 102, 80,GRAY);
    drawBitmap(245, 110, stars, 102, 80,GRAY);
    drawBitmap(245, 190, stars, 102, 80,GRAY);
    drawBitmap(245, 230, stars, 102, 80,GRAY);
    drawBitmap(245, 310, stars, 102, 80,GRAY);
  }
}

int botSize;
int pipeX = 500;
int topLevel;
int botLevel;

// Draws a pair of pipes, with a specific space
void drawPipes(int x, int space) {
  
  tft.fillRect(16, x, botSize - space, 50, RED);
  tft.fillRect(16 + botSize + space, x, 320 - botSize - space, 50, RED);
}

// Deletes the pipes, by drawing the color of the background for day or night
void clearPipesEff(int x, int space, int offset) {
    if (isDay) {
      tft.fillRect(16, x + 50 - offset, botSize - space, offset, DAY);
      tft.fillRect(16 + botSize + space, x + 50 - offset, 320 - botSize - space - 16, offset, DAY);
    } else {
      tft.fillRect(16, x + 50 - offset, botSize - space, offset, NIGHT);
      tft.fillRect(16 + botSize + space, x + 50 - offset, 320 - botSize - space - 16, offset, NIGHT);
    }
}

void drawPipesEff(int x, int space, int offset) {
    tft.fillRect(16, x, botSize - space, offset, RED);
    tft.fillRect(16 + botSize + space, x, 320 - botSize - space - 16, offset, RED);
}

// Draws the day environment
void drawDay() {
  tft.fillScreen(DAY);
  drawPipes(600, 60);

  tft.fillRect(0, 0, 16, 500, GREEN);
  if (!started) {
    drawBitmap(245, 410, sun, 60, 60, ORANGE);
    drawBitmap(235, 70, cloud, 80, 117,WHITE);
    drawBitmap(235, 250, cloud, 80, 117,WHITE);
    drawBitmap(15, 80, building, 150, 150, 0xA513);
    drawBitmap(15, 270, building, 150, 150, 0xA513);
  }  
}

// Function that draws a bitmap image on the LCD
void drawBitmap(int16_t x, int16_t y,const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color) {

  int16_t i, j, byteWidth = (w + 7) / 8;
  uint8_t byte;

  for(j = 0; j < h; j++) {
    for(i=0; i<w; i++) {
      if(i & 7) byte <<= 1;
      else      byte   = pgm_read_byte(bitmap + j * byteWidth + i / 8);
      if(byte & 0x80) tft.drawPixel(x+i, y+j, color);
    }
  }
}

int play_coin = 0;
int coin_index = 0;
int last_buzzed = -1000;

// Plays the current note from the coin melody
void playCoinMelody(int* melody, int notes) {
  if(coin_index < 4) {
   divider = melody[coin_index + 1];
    if (divider > 0) {
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) {
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
      }
 
    if (millis() - last_buzzed > noteDuration) {
      if (coin_index != 0) { 
        noTone(buzzer);
      }
   
      tone(buzzer, melody[coin_index], noteDuration);
      coin_index += 2;
      last_buzzed = millis();
    }
  } else play_coin = 0;
}

bool passedPipe = false;

//Plays the melody when the bird crashes
void playFailMelody(int* melody, int notes) {

  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {
    divider = melody[thisNote + 1];
    if (divider > 0) {
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) {
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
    }
    tone(buzzer, melody[thisNote], noteDuration);
    delay(noteDuration);
    noTone(buzzer);
  }
}

int flappyHeight = 140;
int bufferedPress = 0;

/* Function used to reset the game state, whenever the
player wants to play again */
void initRun() {
  level = 0;
  time_stamp = 0;
  if (isDay) {
    isDay = 0;
  } else isDay = 1;
  bufferedPress = 0;

  randomSeed(153);
  play_coin = 0;
  coin_index = 0;
  last_buzzed = -1000;
  leds = 0;
  updateShiftRegister();
  flappyHeight = 140;
  animationCooldown = 0;
  passedPipe = false;
  flying = 0;
  score = 0;
  pipeX = 500;
  pipe_space = 100;
  if (isDay) {
    drawDay();
  } else {
    drawNight();
  }
  drawFlappy(flappyHeight, 10, flappyColor, 1.2 , 1.2);
  drawStartText();
}

void pin_ISR()
{
  if (flappyColor == BLUE)
    flappyColor = YELLOW;
  else if (flappyColor == YELLOW)
    flappyColor = RED;
  else if (flappyColor == RED)
    flappyColor = CYAN;
  else if (flappyColor == CYAN)
    flappyColor = PURPLE;
  else if (flappyColor == PURPLE)
    flappyColor = BLUE;
}

void setup() {
  Serial.begin(9600);
  pinMode(latchPin, OUTPUT);
  pinMode(dataPin, OUTPUT);  
  pinMode(clockPin, OUTPUT);
  pinMode(2, INPUT_PULLUP);
  pinMode(4, INPUT_PULLUP);
  pinMode(A5, INPUT_PULLUP);
  pinMode(buzzer, OUTPUT);
  pinMode(13, OUTPUT);
  uint16_t ID = tft.readID();
  tft.begin(ID);
  tft.setRotation(0);

  cli();

  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;
  OCR1A = 12499;
  TCCR1B |= (1 << WGM12);
  TCCR1B |= (1 << CS12);
  TIMSK1 |= (1 << OCIE1A);
  sei();
  
  DDRB &= ~(1 << BUTTON);
  PORTB |= (1 << BUTTON);
  PCICR |= (1 << PCIE0);
  PCMSK0 |= (1 << PCINT4);

  randomSeed(153);

  initRun();
}

void loop() {
 if (play_coin)
    playCoinMelody(coin_melody, coin_notes);

    if (lost) {
      high_score = score > high_score ? score : high_score;
      drawGameOver();
      while (lost) {
        if (colorChanged) {
          colorChanged = false;
          clearFlappy(flappyHeight, 10, 1.2 , 1.2);
          drawFlappy(flappyHeight, 10, flappyColor, 1.2 , 1.2);
        }
        if (isPressed()) {
          lost = false;
          started = false;
          initRun();
        }
      }
    }
  if (!lost) {
    animationCooldown -= 1;
    if (started && animationCooldown == 0) {
      clearFlappy(flappyHeight, 10, 1.2 , 1.2);
      if(flying > 0) {
        flappyHeight += 5;
        if (flappyHeight > 300) {
          lost = true;
        }
        flying -= 5;
      } else {
        flappyHeight -= 4;
        if (flappyHeight <= 10) lost = true;
      }
      drawFlappy(flappyHeight, 10, flappyColor, 1.2 , 1.2);
      int offset;
      int new_leds = leds;
      if (passedPipe) {
          passedPipe = false;
        if (level >= 4) {
          new_leds |= 1 << 2;
          offset = 15;
          pipe_space = 50;
        } else if (level == 3) {
          new_leds |= 1 << 3;
          offset = 10;
          pipe_space = 50;
        } else if (level == 2) {
          new_leds |= 1 << 4;
          offset = 15;
        } else if (level == 1) {
          new_leds |= 1 << 5;
          offset = 10;
        } else {
          new_leds |= 1 << 6;
          offset = 5;
        }
        if (new_leds != leds) {
          leds = new_leds;
          updateShiftRegister();
        }
      }
      clearPipesEff(pipeX, pipe_space, offset);

      pipeX -= offset;
      if (pipeX <= -50) {
        pipeX = 500;
        botSize = random(pipe_space, 284);
        botLevel = botSize - pipe_space + 16;
        topLevel = botSize + pipe_space;
        score++;
        passedPipe = true;
        play_coin = 1;
        coin_index = 0;
      }
      if (pipeX <= 40 && (flappyHeight <= botLevel || flappyHeight >= topLevel - 2)) {
        lost = true;
      }
      
      drawPipesEff(pipeX, pipe_space, offset);
      animationCooldown = 10;
    }
    if (bufferedPress || isPressed()) {
      bufferedPress = 0;
      if(started == 0) {
        started = 1;
        botSize = random(20, 284);
        botLevel = botSize - pipe_space + 16;
        topLevel = 16 + botSize + pipe_space;
        if (isDay) {
          drawDay();
        } else {
          drawNight();
        }
      } else {
        flying = 50;
      }
      animationCooldown = 1;
    } else if (!started && colorChanged) {
      colorChanged = 0;
      clearFlappy(flappyHeight, 10, 1.2 , 1.2);
      drawFlappy(flappyHeight, 10, flappyColor, 1.2 , 1.2);
    }
  }
}
